
from ingest.api.ingestapi import IngestApi
from py2neo import Node, Relationship

import re

def convert_to_macrocase(var_name):
    """Converts a variable TO_MACRO_CASE_NAMING_CONVENTION."""
    return ('_').join([a.upper() for a in re.split("([A-Z][^A-Z]*)", var_name) if a])

def flatten(d, parent_key=""):
    """
    Flattens a dict

    Flattens a dictionary extending the keys to include the whole path separated by periods. Also stringifies any lists
    inside the dict. Note dict contained inside lists will get stringified without flattening.
    """

    items = []

    for key, value in d.items():
        new_key = parent_key + "." + key if parent_key else key

        if isinstance(value, dict):
            items.extend(flatten(value, new_key).items())
        # Convert inner lists to strings.
        elif isinstance(value, list):
            items.append((new_key, str(value)))
        else:
            items.append((new_key, value))

    return dict(items)


def main():
    subid = "1869454f-4af2-40bd-bbd3-4e88102be5f6"
    # subid = "b2140c13-c7f7-4b9f-a5ec-5727328c9174"
    ingest = IngestApi("https://api.ingest.archive.data.humancellatlas.org")
    entities = fetch_submission(subid, ingest)
    nodes=get_nodes(entities)
    get_edges(entities, ingest, nodes)


def fetch_submission(subid, ingest):
    # self._logger.debug("fetching ingest submission")

    entities = {}
    fetch_url = f"https://api.ingest.archive.data.humancellatlas.org/submissionEnvelopes/search/findByUuidUuid?uuid={subid}"
    id_field_map = {
        'biomaterials': "biomaterial_core.biomaterial_id",
        'files': "file_core.file_name",
        'processes': "process_core.process_id",
        'projects': "project_core.project_short_name",
        'protocols': "protocol_core.protocol_id",
    }

    for entity_type in ["biomaterials", "files", "processes", "projects", "protocols"]:
        for entity in ingest.get_entities(fetch_url, entity_type):
            properties = flatten(entity['content'])

            new_entity = {
                'properties': properties,
                'labels': [entity['type'].lower()],
                'node_id': properties[id_field_map[entity_type]],
                'links': entity['_links'],
                'uuid': entity['uuid']['uuid'],
            }

            concrete_type = new_entity['properties']['describedBy'].rsplit('/', 1)[1]
            new_entity['labels'].append(concrete_type)

            entities[entity['uuid']['uuid']] = new_entity

    return entities


def get_nodes(entities):
    # self._logger.debug("importing nodes")

    nodes = {}

    for entity_uuid, entity in entities.items():
        node_id = entity['node_id']
        nodes[entity_uuid] = Node(*entity['labels'], **entity['properties'], id=node_id)

        # self._logger.debug(f"({node_id})")

    print(f"imported {len(nodes)} nodes")

    return nodes


def get_edges(entities, ingest_api, nodes):
    # self._logger.debug("importing edges")

    edges = []
    relationship_map = {
        'projects': "projects",
        'protocols': "protocols",
        'inputToProcesses': "processes",
        'derivedByProcesses': "processes",
        'inputBiomaterials': "biomaterials",
        'derivedBiomaterials': "biomaterials",
        'supplementaryFiles': "files",
        'inputFiles': "files",
        'derivedFiles': "files",
    }

    for entity_uuid, entity in entities.items():
        for relationship_type in relationship_map.keys():
            if relationship_type in entity['links']:
                relationships = ingest_api.get_all(
                    entity['links'][relationship_type]['href'],
                    relationship_map[relationship_type]
                )

                for end_entity in relationships:
                    start_node = nodes[entity_uuid]
                    relationship_name = convert_to_macrocase(relationship_type)
                    try:
                        end_node = nodes[end_entity['uuid']['uuid']]
                        edges.append(Relationship(start_node, relationship_name, end_node))

                        # Adding additional relationships to the graphs.
                        if relationship_name == 'INPUT_TO_PROCESSES':
                            edges.append(Relationship(start_node, 'DUMMY_EXPERIMENTAL_DESIGN', end_node))
                        if relationship_name == 'DERIVED_BY_PROCESSES':
                            edges.append(Relationship(end_node, 'DUMMY_EXPERIMENTAL_DESIGN', start_node))

                        print(f"({start_node['id']})-[:{relationship_name}]->({end_node['id']})")

                    except KeyError:
                        print(f"Missing end node at a [{start_node['id']}] entity.")
                        # self._logger.debug(f"Missing end node at a [{start_node['label']}] entity.")

                    # self._logger.debug(f"({start_node['id']})-[:{relationship_name}]->({end_node['id']})")
                    # print(f"({start_node['id']})-[:{relationship_name}]->({end_node['id']})")

    # self._logger.info(f"imported {len(edges)} edges")
    print(f"imported {len(edges)} edges")

    return edges


if __name__ == "__main__":
    main()
